@@ -354,7 +354,7 @@ create_auth (SoupAuthManagerPrivate *priv, SoupMessage *msg)
 	const char *header;
 	SoupAuthClass *auth_class;
 	char *challenge = NULL;
-	SoupAuth *auth;
+	SoupAuth *auth = NULL;
 	int i;
 
 	header = auth_header_for_message (msg);
@@ -364,14 +364,14 @@ create_auth (SoupAuthManagerPrivate *priv, SoupMessage *msg)
 	for (i = priv->auth_types->len - 1; i >= 0; i--) {
 		auth_class = priv->auth_types->pdata[i];
 		challenge = soup_auth_manager_extract_challenge (header, auth_class->scheme_name);
-		if (challenge)
+		if (!challenge)
+			continue;
+		auth = soup_auth_new (G_TYPE_FROM_CLASS (auth_class), msg, challenge);
+		g_free (challenge);
+		if (auth)
 			break;
 	}
-	if (!challenge)
-		return NULL;
 
-	auth = soup_auth_new (G_TYPE_FROM_CLASS (auth_class), msg, challenge);
-	g_free (challenge);
 	return auth;
 }
 
@@ -472,10 +472,21 @@ lookup_auth (SoupAuthManagerPrivate *priv, SoupMessage *msg)
 	if (auth && soup_auth_is_ready (auth, msg))
 		return auth;
 
+	if (soup_message_get_flags (msg) & SOUP_MESSAGE_DO_NOT_USE_AUTH_CACHE)
+		return NULL;
+
 	host = get_auth_host_for_uri (priv, soup_message_get_uri (msg));
 	if (!host->auth_realms && !make_auto_ntlm_auth (priv, host))
 		return NULL;
 
+	/* Cannot change the above '&&' into '||', because make_auto_ntlm_auth() is used
+	 * to populate host->auth_realms when it's not set yet. Even the make_auto_ntlm_auth()
+	 * returns TRUE only if it also populates the host->auth_realms, this extra test
+	 * is required to mute a FORWARD_NULL Coverity Scan warning, which is a false-positive
+	 * here */
+	if (!host->auth_realms)
+		return NULL;
+
 	path = soup_message_get_uri (msg)->path;
 	if (!path)
 		path = "/";
@@ -496,6 +507,9 @@ lookup_proxy_auth (SoupAuthManagerPrivate *priv, SoupMessage *msg)
 	if (auth && soup_auth_is_ready (auth, msg))
 		return auth;
 
+	if (soup_message_get_flags (msg) & SOUP_MESSAGE_DO_NOT_USE_AUTH_CACHE)
+		return NULL;
+
 	return priv->proxy_auth;
 }
 
@@ -625,7 +639,7 @@ auth_got_headers (SoupMessage *msg, gpointer manager)
 	/* If we need to authenticate, try to do it. */
 	authenticate_auth (manager, auth, msg,
 			   prior_auth_failed, FALSE, TRUE);
-	soup_message_set_auth (msg, soup_auth_is_ready (auth, msg) ? auth : NULL);
+	soup_message_set_auth (msg, auth);
 	g_object_unref (auth);
 	g_mutex_unlock (&priv->lock);
 }
@@ -689,7 +703,7 @@ proxy_auth_got_headers (SoupMessage *msg, gpointer manager)
 	/* If we need to authenticate, try to do it. */
 	authenticate_auth (manager, auth, msg,
 			   prior_auth_failed, TRUE, TRUE);
-	soup_message_set_proxy_auth (msg, soup_auth_is_ready (auth, msg) ? auth : NULL);
+	soup_message_set_proxy_auth (msg, auth);
 	g_object_unref (auth);
 	g_mutex_unlock (&priv->lock);
 }
