@@ -39,17 +39,12 @@ soup_headers_parse (const char *str, int len, SoupMessageHeaders *dest)
 	const char *headers_start;
 	char *headers_copy, *name, *name_end, *value, *value_end;
 	char *eol, *sol, *p;
+	gsize copy_len;
 	gboolean success = FALSE;
 
 	g_return_val_if_fail (str != NULL, FALSE);
 	g_return_val_if_fail (dest != NULL, FALSE);
 
-	/* RFC 2616 does allow NUL bytes in the headers, but httpbis
-	 * is changing that, and we can't deal with them anyway.
-	 */
-	if (memchr (str, '\0', len))
-		return FALSE;
-
 	/* As per RFC 2616 section 19.3, we treat '\n' as the
 	 * line terminator, and '\r', if it appears, merely as
 	 * ignorable trailing whitespace.
@@ -59,14 +54,28 @@ soup_headers_parse (const char *str, int len, SoupMessageHeaders *dest)
 	headers_start = memchr (str, '\n', len);
 	if (!headers_start)
 		return FALSE;
+	/* No '\0's in the Request-Line / Status-Line */
+	if (memchr (str, '\0', headers_start - str))
+		return FALSE;
 
 	/* We work on a copy of the headers, which we can write '\0's
 	 * into, so that we don't have to individually g_strndup and
 	 * then g_free each header name and value.
 	 */
-	headers_copy = g_strndup (headers_start, len - (headers_start - str));
+	copy_len = len - (headers_start - str);
+	headers_copy = g_malloc (copy_len + 1);
+	memcpy (headers_copy, headers_start, copy_len);
+	headers_copy[copy_len] = '\0';
 	value_end = headers_copy;
 
+	/* There shouldn't be any '\0's in the headers already, but
+	 * this is the web we're talking about.
+	 */
+	while ((p = memchr (headers_copy, '\0', copy_len))) {
+		memmove (p, p + 1, copy_len - (p - headers_copy));
+		copy_len--;
+	}
+
 	while (*(value_end + 1)) {
 		name = value_end + 1;
 		name_end = strchr (name, ':');
@@ -314,7 +323,7 @@ soup_headers_parse_status_line (const char       *status_line,
 	if (code_end != code_start + 3)
 		return FALSE;
 	code = atoi (code_start);
-	if (code < 100 || code > 599)
+	if (code < 100 || code > 999)
 		return FALSE;
 	if (status_code)
 		*status_code = code;
@@ -621,11 +630,13 @@ gboolean
 soup_header_contains (const char *header, const char *token)
 {
 	const char *end;
-	guint len = strlen (token);
+	guint len;
 
 	g_return_val_if_fail (header != NULL, FALSE);
 	g_return_val_if_fail (token != NULL, FALSE);
 
+	len = strlen (token);
+
 	header = skip_delims (header, ',');
 	while (*header) {
 		end = skip_item (header, ',');
@@ -699,12 +710,12 @@ decode_rfc5987 (char *encoded_string)
 }
 
 static GHashTable *
-parse_param_list (const char *header, char delim)
+parse_param_list (const char *header, char delim, gboolean strict)
 {
 	GHashTable *params;
 	GSList *list, *iter;
 	char *item, *eq, *name_end, *value;
-	gboolean override;
+	gboolean override, duplicated;
 
 	params = g_hash_table_new_full (soup_str_case_hash, 
 					soup_str_case_equal,
@@ -740,7 +751,14 @@ parse_param_list (const char *header, char delim)
 		} else
 			value = NULL;
 
-		if (override || !g_hash_table_lookup (params, item))
+		duplicated = g_hash_table_lookup_extended (params, item, NULL, NULL);
+
+		if (strict && duplicated) {
+			soup_header_free_param_list (params);
+			params = NULL;
+			g_free (item);
+			break;
+		} else if (override || !duplicated)
 			g_hash_table_replace (params, item, value);
 		else
 			g_free (item);
@@ -773,7 +791,7 @@ soup_header_parse_param_list (const char *header)
 {
 	g_return_val_if_fail (header != NULL, NULL);
 
-	return parse_param_list (header, ',');
+	return parse_param_list (header, ',', FALSE);
 }
 
 /**
@@ -801,7 +819,61 @@ soup_header_parse_semi_param_list (const char *header)
 {
 	g_return_val_if_fail (header != NULL, NULL);
 
-	return parse_param_list (header, ';');
+	return parse_param_list (header, ';', FALSE);
+}
+
+/**
+ * soup_header_parse_param_list_strict:
+ * @header: a header value
+ *
+ * A strict version of soup_header_parse_param_list()
+ * that bails out if there are duplicate parameters.
+ * Note that this function will treat RFC5987-encoded
+ * parameters as duplicated if an ASCII version is also
+ * present. For header fields that might contain
+ * RFC5987-encoded parameters, use
+ * soup_header_parse_param_list() instead.
+ *
+ * Return value: (element-type utf8 utf8) (transfer full) (nullable):
+ * a #GHashTable of list elements, which can be freed with
+ * soup_header_free_param_list() or %NULL if there are duplicate
+ * elements.
+ *
+ * Since: 2.66
+ **/
+GHashTable *
+soup_header_parse_param_list_strict (const char *header)
+{
+	g_return_val_if_fail (header != NULL, NULL);
+
+	return parse_param_list (header, ',', TRUE);
+}
+
+/**
+ * soup_header_parse_semi_param_list_strict:
+ * @header: a header value
+ *
+ * A strict version of soup_header_parse_semi_param_list()
+ * that bails out if there are duplicate parameters.
+ * Note that this function will treat RFC5987-encoded
+ * parameters as duplicated if an ASCII version is also
+ * present. For header fields that might contain
+ * RFC5987-encoded parameters, use
+ * soup_header_parse_semi_param_list() instead.
+ *
+ * Return value: (element-type utf8 utf8) (transfer full) (nullable):
+ * a #GHashTable of list elements, which can be freed with
+ * soup_header_free_param_list() or %NULL if there are duplicate
+ * elements.
+ *
+ * Since: 2.66
+ **/
+GHashTable *
+soup_header_parse_semi_param_list_strict (const char *header)
+{
+	g_return_val_if_fail (header != NULL, NULL);
+
+	return parse_param_list (header, ';', TRUE);
 }
 
 /**
