@@ -148,8 +148,6 @@ struct _SoupWebsocketConnectionPrivate {
 };
 
 #define MAX_INCOMING_PAYLOAD_SIZE_DEFAULT   128 * 1024
-#define READ_BUFFER_SIZE 1024
-#define MASK_LENGTH 4
 
 G_DEFINE_TYPE_WITH_PRIVATE (SoupWebsocketConnection, soup_websocket_connection, G_TYPE_OBJECT)
 
@@ -158,11 +156,6 @@ static void queue_frame (SoupWebsocketConnection *self, SoupWebsocketQueueFlags
 
 static void protocol_error_and_close (SoupWebsocketConnection *self);
 
-static gboolean on_web_socket_input (GObject *pollable_stream,
-				     gpointer user_data);
-static gboolean on_web_socket_output (GObject *pollable_stream,
-				      gpointer user_data);
-
 /* Code below is based on g_utf8_validate() implementation,
  * but handling NULL characters as valid, as expected by
  * WebSockets and compliant with RFC 3629.
@@ -291,20 +284,7 @@ on_iostream_closed (GObject *source,
 }
 
 static void
-soup_websocket_connection_start_input_source (SoupWebsocketConnection *self)
-{
-	SoupWebsocketConnectionPrivate *pv = self->pv;
-
-	if (pv->input_source)
-		return;
-
-	pv->input_source = g_pollable_input_stream_create_source (pv->input, NULL);
-	g_source_set_callback (pv->input_source, (GSourceFunc)on_web_socket_input, self, NULL);
-	g_source_attach (pv->input_source, pv->main_context);
-}
-
-static void
-soup_websocket_connection_stop_input_source (SoupWebsocketConnection *self)
+stop_input (SoupWebsocketConnection *self)
 {
 	SoupWebsocketConnectionPrivate *pv = self->pv;
 
@@ -317,20 +297,7 @@ soup_websocket_connection_stop_input_source (SoupWebsocketConnection *self)
 }
 
 static void
-soup_websocket_connection_start_output_source (SoupWebsocketConnection *self)
-{
-	SoupWebsocketConnectionPrivate *pv = self->pv;
-
-	if (pv->output_source)
-		return;
-
-	pv->output_source = g_pollable_output_stream_create_source (pv->output, NULL);
-	g_source_set_callback (pv->output_source, (GSourceFunc)on_web_socket_output, self, NULL);
-	g_source_attach (pv->output_source, pv->main_context);
-}
-
-static void
-soup_websocket_connection_stop_output_source (SoupWebsocketConnection *self)
+stop_output (SoupWebsocketConnection *self)
 {
 	SoupWebsocketConnectionPrivate *pv = self->pv;
 
@@ -375,8 +342,8 @@ close_io_stream (SoupWebsocketConnection *self)
 	close_io_stop_timeout (self);
 
 	if (!pv->io_closing) {
-		soup_websocket_connection_stop_input_source (self);
-		soup_websocket_connection_stop_output_source (self);
+		stop_input (self);
+		stop_output (self);
 		pv->io_closing = TRUE;
 		g_debug ("closing io stream");
 		g_io_stream_close_async (pv->io_stream, G_PRIORITY_DEFAULT,
@@ -394,7 +361,7 @@ shutdown_wr_io_stream (SoupWebsocketConnection *self)
 	GIOStream *base_iostream;
 	GError *error = NULL;
 
-	soup_websocket_connection_stop_output_source (self);
+	stop_output (self);
 
 	base_iostream = SOUP_IS_IO_STREAM (pv->io_stream) ?
 		soup_io_stream_get_base_iostream (SOUP_IO_STREAM (pv->io_stream)) :
@@ -464,7 +431,8 @@ send_message (SoupWebsocketConnection *self,
 	GByteArray *bytes;
 	gsize frame_len;
 	guint8 *outer;
-	guint8 mask_offset;
+	guint8 *mask = 0;
+	guint8 *at;
 
 	if (!(soup_websocket_connection_get_state (self) == SOUP_WEBSOCKET_STATE_OPEN)) {
 		g_debug ("Ignoring message since the connection is closed or is closing");
@@ -480,7 +448,6 @@ send_message (SoupWebsocketConnection *self,
 		if (length > 125) {
 			g_warning ("WebSocket control message payload exceeds size limit");
 			protocol_error_and_close (self);
-			g_byte_array_free (bytes, TRUE);
 			return;
 		}
 
@@ -518,15 +485,16 @@ send_message (SoupWebsocketConnection *self,
 	if (self->pv->connection_type == SOUP_WEBSOCKET_CONNECTION_CLIENT) {
 		guint32 rnd = g_random_int ();
 		outer[1] |= 0x80;
-		mask_offset = bytes->len;
-		memcpy (outer + mask_offset, &rnd, sizeof (rnd));
-		bytes->len += MASK_LENGTH;
+		mask = outer + bytes->len;
+		memcpy (mask, &rnd, sizeof (rnd));
+		bytes->len += 4;
 	}
 
+	at = bytes->data + bytes->len;
 	g_byte_array_append (bytes, data, length);
 
 	if (self->pv->connection_type == SOUP_WEBSOCKET_CONNECTION_CLIENT)
-		xor_with_mask (bytes->data + mask_offset, bytes->data + mask_offset + MASK_LENGTH, length);
+		xor_with_mask (mask, at, length);
 
 	frame_len = bytes->len;
 	queue_frame (self, flags, g_byte_array_free (bytes, FALSE),
@@ -650,6 +618,9 @@ too_big_error_and_close (SoupWebsocketConnection *self,
 		 self->pv->connection_type == SOUP_WEBSOCKET_CONNECTION_SERVER ? "server" : "client",
 	         payload_len, self->pv->max_incoming_payload_size);
 	emit_error_and_close (self, error, TRUE);
+
+	/* The input is in an invalid state now */
+	stop_input (self);
 }
 
 static void
@@ -689,10 +660,6 @@ close_connection (SoupWebsocketConnection *self,
 			         code);
 		}
 		break;
-	case SOUP_WEBSOCKET_CLOSE_NO_STATUS:
-		/* This is special case to send a close message with no body */
-		code = 0;
-		break;
 	default:
 		if (code < 3000) {
 			g_debug ("Wrong closing code %d received", code);
@@ -728,7 +695,6 @@ receive_close (SoupWebsocketConnection *self,
 	switch (len) {
 	case 0:
 		/* Send a clean close when having an empty payload */
-		pv->peer_close_code = SOUP_WEBSOCKET_CLOSE_NO_STATUS;
 		close_connection (self, 1000, NULL);
 		return;
 	case 1:
@@ -806,9 +772,6 @@ process_contents (SoupWebsocketConnection *self,
 	SoupWebsocketConnectionPrivate *pv = self->pv;
 	GBytes *message;
 
-	if (pv->close_sent && pv->close_received)
-		return;
-
 	if (control) {
 		/* Control frames must never be fragmented */
 		if (!fin) {
@@ -1059,31 +1022,32 @@ process_incoming (SoupWebsocketConnection *self)
 		;
 }
 
-static void
-soup_websocket_connection_read (SoupWebsocketConnection *self)
+static gboolean
+on_web_socket_input (GObject *pollable_stream,
+		     gpointer user_data)
 {
+	SoupWebsocketConnection *self = SOUP_WEBSOCKET_CONNECTION (user_data);
 	SoupWebsocketConnectionPrivate *pv = self->pv;
 	GError *error = NULL;
 	gboolean end = FALSE;
 	gssize count;
 	gsize len;
 
-	soup_websocket_connection_stop_input_source (self);
-
 	do {
 		len = pv->incoming->len;
-		g_byte_array_set_size (pv->incoming, len + READ_BUFFER_SIZE);
+		g_byte_array_set_size (pv->incoming, len + 1024);
 
 		count = g_pollable_input_stream_read_nonblocking (pv->input,
 								  pv->incoming->data + len,
-								  READ_BUFFER_SIZE, NULL, &error);
+								  1024, NULL, &error);
+
 		if (count < 0) {
 			if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_WOULD_BLOCK)) {
 				g_error_free (error);
 				count = 0;
 			} else {
 				emit_error_and_close (self, error, TRUE);
-				return;
+				return TRUE;
 			}
 		} else if (count == 0) {
 			end = TRUE;
@@ -1103,25 +1067,16 @@ soup_websocket_connection_read (SoupWebsocketConnection *self)
 		}
 
 		close_io_stream (self);
-		return;
 	}
 
-	if (!pv->io_closing)
-		soup_websocket_connection_start_input_source (self);
+	return TRUE;
 }
 
 static gboolean
-on_web_socket_input (GObject *pollable_stream,
-		     gpointer user_data)
-{
-	soup_websocket_connection_read (SOUP_WEBSOCKET_CONNECTION (user_data));
-
-	return G_SOURCE_REMOVE;
-}
-
-static void
-soup_websocket_connection_write (SoupWebsocketConnection *self)
+on_web_socket_output (GObject *pollable_stream,
+		      gpointer user_data)
 {
+	SoupWebsocketConnection *self = SOUP_WEBSOCKET_CONNECTION (user_data);
 	SoupWebsocketConnectionPrivate *pv = self->pv;
 	const guint8 *data;
 	GError *error = NULL;
@@ -1129,18 +1084,19 @@ soup_websocket_connection_write (SoupWebsocketConnection *self)
 	gssize count;
 	gsize len;
 
-	soup_websocket_connection_stop_output_source (self);
-
 	if (soup_websocket_connection_get_state (self) == SOUP_WEBSOCKET_STATE_CLOSED) {
 		g_debug ("Ignoring message since the connection is closed");
-		return;
+		stop_output (self);
+		return TRUE;
 	}
 
 	frame = g_queue_peek_head (&pv->outgoing);
 
 	/* No more frames to send */
-	if (frame == NULL)
-		return;
+	if (frame == NULL) {
+		stop_output (self);
+		return TRUE;
+	}
 
 	data = g_bytes_get_data (frame->data, &len);
 	g_assert (len > 0);
@@ -1160,7 +1116,7 @@ soup_websocket_connection_write (SoupWebsocketConnection *self)
 			frame->pending = TRUE;
 		} else {
 			emit_error_and_close (self, error, TRUE);
-			return;
+			return FALSE;
 		}
 	}
 
@@ -1178,21 +1134,23 @@ soup_websocket_connection_write (SoupWebsocketConnection *self)
 			}
 		}
 		frame_free (frame);
-
-		if (g_queue_is_empty (&pv->outgoing))
-			return;
 	}
 
-	soup_websocket_connection_start_output_source (self);
+	return TRUE;
 }
 
-static gboolean
-on_web_socket_output (GObject *pollable_stream,
-		      gpointer user_data)
+static void
+start_output (SoupWebsocketConnection *self)
 {
-	soup_websocket_connection_write (SOUP_WEBSOCKET_CONNECTION (user_data));
+	SoupWebsocketConnectionPrivate *pv = self->pv;
 
-	return G_SOURCE_REMOVE;
+	if (pv->output_source)
+		return;
+
+	g_debug ("starting output source");
+	pv->output_source = g_pollable_output_stream_create_source (pv->output, NULL);
+	g_source_set_callback (pv->output_source, (GSourceFunc)on_web_socket_output, self, NULL);
+	g_source_attach (pv->output_source, pv->main_context);
 }
 
 static void
@@ -1233,7 +1191,7 @@ queue_frame (SoupWebsocketConnection *self,
 		g_queue_push_tail (&pv->outgoing, frame);
 	}
 
-	soup_websocket_connection_write (self);
+	start_output (self);
 }
 
 static void
@@ -1258,7 +1216,9 @@ soup_websocket_connection_constructed (GObject *object)
 	pv->output = G_POLLABLE_OUTPUT_STREAM (os);
 	g_return_if_fail (g_pollable_output_stream_can_poll (pv->output));
 
-	soup_websocket_connection_start_input_source (self);
+	pv->input_source = g_pollable_input_stream_create_source (pv->input, NULL);
+	g_source_set_callback (pv->input_source, (GSourceFunc)on_web_socket_input, self, NULL);
+	g_source_attach (pv->input_source, pv->main_context);
 }
 
 static void
@@ -1905,6 +1865,38 @@ soup_websocket_connection_send_binary (SoupWebsocketConnection *self,
 	send_message (self, SOUP_WEBSOCKET_QUEUE_NORMAL, 0x02, data, length);
 }
 
+/**
+ * soup_websocket_connection_send_message:
+ * @self: the WebSocket
+ * @type: the type of message contents
+ * @message: the message data as #GBytes
+ *
+ * Send a message of the given @type to the peer. Note that this method,
+ * allows to send text messages containing %NULL characters.
+ *
+ * The message is queued to be sent and will be sent when the main loop
+ * is run.
+ *
+ * Since: 2.68
+ */
+void
+soup_websocket_connection_send_message (SoupWebsocketConnection *self,
+                                        SoupWebsocketDataType type,
+                                        GBytes *message)
+{
+        gconstpointer data;
+        gsize length;
+
+        g_return_if_fail (SOUP_IS_WEBSOCKET_CONNECTION (self));
+        g_return_if_fail (soup_websocket_connection_get_state (self) == SOUP_WEBSOCKET_STATE_OPEN);
+        g_return_if_fail (message != NULL);
+
+        data = g_bytes_get_data (message, &length);
+        g_return_if_fail (type != SOUP_WEBSOCKET_DATA_TEXT || utf8_validate ((const char *)data, length));
+
+        send_message (self, SOUP_WEBSOCKET_QUEUE_NORMAL, (int)type, data, length);
+}
+
 /**
  * soup_websocket_connection_close:
  * @self: the WebSocket
@@ -1918,8 +1910,6 @@ soup_websocket_connection_send_binary (SoupWebsocketConnection *self,
  * main loop runs.
  *
  * The @code and @data are sent to the peer along with the close request.
- * If @code is %SOUP_WEBSOCKET_CLOSE_NO_STATUS a close message with no body
- * (without code and data) is sent.
  * Note that the @data must be UTF-8 valid.
  *
  * Since: 2.50
@@ -1935,7 +1925,8 @@ soup_websocket_connection_close (SoupWebsocketConnection *self,
 	pv = self->pv;
 	g_return_if_fail (!pv->close_sent);
 
-	g_return_if_fail (code != SOUP_WEBSOCKET_CLOSE_ABNORMAL &&
+	g_return_if_fail (code != SOUP_WEBSOCKET_CLOSE_NO_STATUS &&
+			  code != SOUP_WEBSOCKET_CLOSE_ABNORMAL &&
 			  code != SOUP_WEBSOCKET_CLOSE_TLS_HANDSHAKE);
 	if (pv->connection_type == SOUP_WEBSOCKET_CONNECTION_SERVER)
 		g_return_if_fail (code != SOUP_WEBSOCKET_CLOSE_NO_EXTENSION);
